// $Id: avistream.cpp 353 2015-09-19 16:06:32Z you $
/*
 * Copyright (c) 2007-2015 akamoz.jp
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <assert.h>
#include <windows.h>
#include <vfw.h>
#include <akamoz-image.h>

namespace akamoz {

AVIStream::AVIStream()
{
	gAVIiniter.init();
	_init();
}

template <typename TC>
AVIStream::AVIStream(const TC *pszfn)
{
	_init();
	setup(pszfn);
}

template AVIStream::AVIStream(LPCSTR pszfn);
template AVIStream::AVIStream(LPCWSTR pszfn);

AVIStream::~AVIStream()
{
	_cleanup();
}

void AVIStream::_init()
{
	mbOpened = 0;
}

void AVIStream::_cleanup()
{
	if (mbOpened)
		AVIStreamRelease(mpavistr);
}

template <typename TC>
int AVIStream::setup(const TC *pszfn)
{
	if (pszfn == NULL)
		return AKAMOZ_ERR(-1);
	if (
		aviStreamOpenFromFile(&mpavistr, pszfn, streamtypeVIDEO, 0,
			OF_READ, NULL) != 0
	)
		return AKAMOZ_ERR(-1);
	mbOpened = 1;
	return 0;
}

template int AVIStream::setup(LPCSTR pszfn);
template int AVIStream::setup(LPCWSTR pszfn);

int AVIStream::cleanup()
{
	_cleanup();
	_init();
	return 0;
}

int AVIStream::getInfo(AVISTREAMINFO *pavisi)
{
	if (pavisi == NULL)
		return AKAMOZ_ERR(-1);
	if (!isReady())
		return AKAMOZ_ERR(-1);
	if (AVIStreamInfo(mpavistr, pavisi, sizeof(*pavisi)) != 0)
		return AKAMOZ_ERR(-1);
	return 0;
}

int AVIStream::getFormat(BITMAPINFOHEADER *pbmih, long lPos)
{
	if (pbmih == NULL)
		return AKAMOZ_ERR(-1);
	if (!isReady())
		return AKAMOZ_ERR(-1);
	LONG lSize = sizeof(*pbmih);
	if (AVIStreamReadFormat(mpavistr, lPos, pbmih, &lSize) != 0)
		return AKAMOZ_ERR(-1);
	return 0;
}

int AVIStream::getFormat(DataBuffer *pdb, long lPos)
{
	if (pdb == NULL)
		return AKAMOZ_ERR(-1);
	if (!isReady())
		return AKAMOZ_ERR(-1);
	LONG lSize = -1;
	if (AVIStreamFormatSize(mpavistr, lPos, &lSize) != 0)
		return AKAMOZ_ERR(-1);
	if (lSize < 0)
		return AKAMOZ_ERR(-1);
	if (pdb->reuse(lSize) < 0)
		return AKAMOZ_ERR(-1);
	if (AVIStreamReadFormat(mpavistr, lPos, (BYTE *)*pdb, &lSize) != 0)
		return AKAMOZ_ERR(-1);
	if (lSize < 0)
		return AKAMOZ_ERR(-1);
	return 0;
}

int AVIStream::getFrameSize(long lPos)
{
	assert(isReady());
	long lSize;
	if (AVIStreamSampleSize(mpavistr, lPos, &lSize) != 0)
		return AKAMOZ_ERR(-1);
	return lSize;
}

int AVIStream::getFrame(ImageBuffer *pimg, long lPos)
{
	if (pimg == NULL)
		return AKAMOZ_ERR(-1);
	if (lPos < 0)
		return AKAMOZ_ERR(-1);
	if (!isReady())
		return AKAMOZ_ERR(-1);
	BITMAPINFOHEADER bmih;
	if (getFormat(&bmih, lPos) < 0)
		return AKAMOZ_ERR(-1);
	if (bmih.biCompression != pimg->getFormat())
		return AKAMOZ_ERR(-1);
	if ((bmih.biCompression == BI_RGB) && (bmih.biBitCount != 24))
		return AKAMOZ_ERR(-1);
	int nWidth = bmih.biWidth;
	if (nWidth < 0)
		return AKAMOZ_ERR(-1);
	int nHeight = bmih.biHeight;
	if (nHeight < 0)
		return AKAMOZ_ERR(-1);	// top-down DIB
	if (pimg->setProperty(nWidth, nHeight) < 0)
		return AKAMOZ_ERR(-1);
	long lSize = -1;
	if (AVIStreamSampleSize(mpavistr, lPos, &lSize) != 0)
		return AKAMOZ_ERR(-1);
	int nBufSize = pimg->getImageSize();
	if (nBufSize < lSize)
		return AKAMOZ_ERR(-1);
	BYTE *pb = pimg->getPixelBuffer();
	if (pb == NULL)
		return AKAMOZ_ERR(-1);
	long lFrames = 0;
	if (AVIStreamRead(mpavistr, lPos, 1, pb, nBufSize, NULL, &lFrames) != 0)
		return AKAMOZ_ERR(-1);
	if (lFrames < 0)
		return AKAMOZ_ERR(-1);
	if (lFrames > 1)
		return AKAMOZ_ERR(-1);
	return lFrames;
}

int AVIStream::getFrame(DataBuffer *pdb, long lPos)
{
	assert(pdb != NULL);
	assert(lPos >= 0);
	assert(isReady());
	long lSize = getFrameSize(lPos);
	if (lSize < 0)
		return AKAMOZ_ERR(-1);
	if (pdb->reuse(lSize) < 0)
		return AKAMOZ_ERR(-1);
	BYTE *pb = *pdb;
	if (AVIStreamRead(mpavistr, lPos, 1, pb, pdb->size(), &lSize, NULL) != 0)
		return AKAMOZ_ERR(-1);
	return lSize;
}

};
