// $Id: img-yuy2.cpp 358 2015-09-20 10:00:06Z you $
// ImageBufferYUY2: ImageBuffer for YUY2 color space.
/*
 * Copyright (c) 2007-2015 akamoz.jp
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <assert.h>
#include <akamoz-image.h>

namespace akamoz {

ImageBufferYUY2::ImageBufferYUY2()
{
	_init();
}

int ImageBufferYUY2::_calcLineSize(int nWidth)
{
	int nW2 = nWidth + 1;
	if (nW2 < 0)
		return AKAMOZ_ERR(-1);
	nW2 /= 2;
	int nLineSize = nW2 * 4;
	if (nLineSize / 4 != nW2)
		return AKAMOZ_ERR(-1);
	return nLineSize;
}

int ImageBufferYUY2::setimg(const ImageBuffer *pimg)
{
	if (pimg == NULL)
		return AKAMOZ_ERR(-1);
	int nWidth = pimg->getWidth();
	if (nWidth < 1)
		return AKAMOZ_ERR(-1);
	int nHeight = pimg->getHeight();
	if (nHeight < 1)
		return AKAMOZ_ERR(-1);
	if (setProperty(nWidth, nHeight) < 0)
		return AKAMOZ_ERR(-1);

	BYTE *pbLineYUV = getPixelBuffer();
	if (pbLineYUV == NULL)
		return AKAMOZ_ERR(-1);
	const BYTE *pbLineRGB = pimg->getPixelBuffer();
	if (pbLineRGB == NULL)
		return AKAMOZ_ERR(-1);
	int nLineSizeRGB = pimg->getLineSize();
	if (nLineSizeRGB < 0)
		return AKAMOZ_ERR(-1);

	// ITU-R BT.601
	// Ya = +0.299Ra +0.587Ga +0.114Ba
	// Cb = 0.564(Ba - Ya) = -0.169Ra -0.331Ga +0.500Ba
	// Cr = 0.713(Ra - Ya) = +0.500Ra -0.419Ga -0.081Ba
	// Yd = 219Ya + 16
	// U = 224Cb + 128
	// V = 224Cr + 128

	// Ya(0.0 -> 1.0) -> Yd(16 -> 235, 220steps)
	// Below is a little bit different from ITU-R BT.601
	//    = 220(+0.299Ra +0.587Ga +0.114Ba) + 16
	//    = 220(+0.299(R+0.5)/256 +0.587(G+0.5)/256 +0.114(B+0.5)/256) + 16
	//    = 220(+0.299(2R+1) +0.587(2G+1) +0.114(2B+1))/512 + 16
	//	  = (66R + 129G + 25B + 110) / 256 + 16
	//	  = (66R + 129G + 25B + 110 + 0x1000) / 256
	// Cb(-0.5 -> 0.5) -> U(16 -> 240, 225steps)
	//	base expression is: U = 224Cb + 128
	//	but use: U = floor(225Cb + 128.5)
	// U = 225Cb + 128.5
	//   = 225(-0.169Ra -0.331Ga +0.500Ba) + 128.5
	//   = 225(-0.169(R+0.5)/256 -0.331(G+0.5)/256 +0.500(B+0.5)/256) + 128.5
	//   = 225(-0.169(R+0.5)/256 -0.331(G+0.5)/256 +0.500(B+0.5)/256) + 128.5
	//   = 225(-0.169(R+0.5) -0.331(G+0.5) +0.500(B+0.5))/256 + 128.5
	//   = 225(-0.169(R+0.5) -0.331(G+0.5) +0.500(B+0.5) + 0x8080)/256
	//   = 225(-0.169R -0.331G +0.500B + 0.5(-0.169-0.331+0.500) + 0x8080)/256
	//   = 225(-0.169R -0.331G +0.500B + 0.5(0.000) + 0x8080)/256
	//   = 225(-0.169R -0.331G +0.500B + 0x8080)/256
	//   = (-(76R + 149G)/2 +225B/2 + 0x8080)/256
	// V = 225Cr + 128.5
	//   = 225(+0.500Ra -0.419Ga -0.081Ba) + 128.5
	//   = 225(+0.500(R+0.5)/256 -0.419(G+0.5)/256 -0.081(B+0.5)/256) + 128.5
	//   = 225(+0.500(R+0.5) -0.419(G+0.5) -0.081(B+0.5))/256 + 128.5
	//   = 225(+0.500(R+0.5) -0.419(G+0.5) -0.081(B+0.5) + 0x8080)/256
	//   = 225(+0.500R -0.419G -0.081B +0.5(0.500-0.419-0.081) + 0x8080)/256
	//   = 225(+0.500R -0.419G -0.081B +0.5(0.000) + 0x8080)/256
	//   = 225(+0.500R -0.419G -0.081B + 0x8080)/256
	//   = 225(+1.000R/2 -(0.838G + 0.162B)/2 + 0x8080)/256
	//   = (225R/2 -(189G + 36B)/2 + 0x8080)/256

	for (int y = 0; y < mnHeight; y++) {
		const BYTE *pbRGB = pbLineRGB;
		BYTE *pbYUV = pbLineYUV;
		for (int x = 0; x < mnWidth; x+= 2) {
			pbYUV[0] = calcY(pbRGB[2], pbRGB[1], pbRGB[0]);
			pbYUV[1] = calcU(pbRGB[2], pbRGB[1], pbRGB[0]);
			pbYUV[2] = calcY(pbRGB[5], pbRGB[4], pbRGB[3]);
			pbYUV[3] = calcV(pbRGB[2], pbRGB[1], pbRGB[0]);
			pbRGB += 6;
			pbYUV += 4;
		}
		pbLineRGB += nLineSizeRGB;
		pbLineYUV += mnLineSize;
	}

	return 0;
}

// ITU-R BT.601
// Ya = +0.299Ra +0.587Ga +0.114Ba
// Cb = 0.564(Ba - Ya) = -0.169Ra -0.331Ga +0.500Ba
// Cr = 0.713(Ra - Ya) = +0.500Ra -0.419Ga -0.081Ba
// Ba - Ya = Cb / 0.564
// Ra - Ya = Cr / 0.713

// Ba = Cb / 0.564 + Ya
// Ra = Cr / 0.713 + Ya

// U, V: 16 -> 240 -> Cb, Cr: -0.5 -> 0.5
//	(U-16+0.5)/225 - 0.5
//	= (U-15.5)/225 - 0.5
//	= (U-15.5-112.5)/225
//	= (U-128)/225
// Ya(0.0 -> 1.0) -> Yd(16 -> 235, 220steps)
//	(Yd-15.5)/220

// Ga = (Ya - 0.299Ra - 0.114Ba) / 0.587
//    = (Ya - 0.299(Cr / 0.713 + Ya) - 0.114(Cb / 0.564 + Ya)) / 0.587
//    = (Ya - 0.299Cr/0.713 - 0.299Ya - 0.114Cb/0.564 - 0.114Ya) / 0.587
//    = (Ya - 0.413Ya - 0.299Cr/0.713 - 0.114Cb/0.564 - 0.114Ya) / 0.587
//    = (0.587Ya - 0.299Cr/0.713 - 0.114Cb/0.564) / 0.587
//    = Ya - 0.299Cr/0.713/0.587 - 0.114Cb/0.564/0.587
//    = ((Yd-15.5)/220) - 0.299((V-128)/225)/0.713/0.587 - 0.114((U-128)/225)/0.564/0.587
//    = ((Yd-15.5)/220) - 0.299(V-128)/225/0.713/0.587 - 0.114(U-128)/225/0.564/0.587
// Gd = Ga * 256
//    = 256*(Yd-15.5)/220 - 256*0.299(V-128)/225/0.713/0.587 - 256*0.114(U-128)/225/0.564/0.587

// B = Ba * 256 = (Cb / 0.564 + Ya) * 256
//   = ((U-128)/225/0.564 + (Yd-15.5)/220) * 256
// ( = 0.6417(U-128) + 1.1636(Yd-15.5) )

// R = Ra * 256 = (Cr/0.713 + Ya) * 256
//   = ((V-128)/225/0.713 + (Yd-15.5)/220) * 256
// ( = 1.4080(V-128) + 1.1636(Yd-15.5) )

void ImageBufferYUY2::calcYUVtoRGB(int nY, int nU, int nV, BYTE *pRGB)
{
	nY = (nY * 2 - 31) * 5760;
	nU -= 128;
	nV -= 128;
	pRGB[2] = clip256((nY + 15792*nV) / 9900);
	pRGB[1] = clip256((nY - 3876*nU - 8044*nV) / 9900);
	pRGB[0] = clip256((nY + 19960*nU) / 9900);
}

int ImageBufferYUY2::convertToRGB(ImageBuffer *pimgRGB) const
{
	if (!isReady())
		return AKAMOZ_ERR(-1);
	const BYTE *pbLineYUV = getPixelBuffer();
	assert(pbLineYUV != NULL);

	pimgRGB->setProperty(mnWidth, mnHeight);
	if (!pimgRGB->isReady())
		return AKAMOZ_ERR(-1);
	int nLineSizeRGB = pimgRGB->getLineSize();
	assert(nLineSizeRGB > 0);
	BYTE *pbLineRGB = pimgRGB->getPixelBuffer();
	assert(pbLineRGB != NULL);

	assert(mnWidth > 0);
	assert(mnHeight > 0);
	if (mnWidth <= 2) {
		// UV is one sample per line. No interpolation.
		for (int y = 0; y < mnHeight; y++) {
			const BYTE *pbYUV = pbLineYUV;
			BYTE *pbRGB = pbLineRGB;
			calcYUVtoRGB(pbYUV[0], pbYUV[1], pbYUV[3], pbRGB);
			if (mnWidth == 2)
				calcYUVtoRGB(pbYUV[2], pbYUV[1], pbYUV[3], pbRGB + 3);
			pbLineYUV += mnLineSize;
			pbLineRGB += nLineSizeRGB;
		}
	}
	else if (mnWidth <= 4) {
		// UV is two samples per line.
		for (int y = 0; y < mnHeight; y++) {
			const BYTE *pbYUV = pbLineYUV;
			BYTE *pbRGB = pbLineRGB;
			// x = 0
			calcYUVtoRGB(pbYUV[0], pbYUV[1], pbYUV[3], pbRGB);
			// x = 1, UV is just a mean.
			int nU1 = (pbYUV[1] + pbYUV[5] + 1) / 2;
			int nV1 = (pbYUV[3] + pbYUV[7] + 1) / 2;
			calcYUVtoRGB(pbYUV[2], nU1, nV1, pbRGB + 3);
			// x = 2
			calcYUVtoRGB(pbYUV[4], pbYUV[5], pbYUV[7], pbRGB + 6);
			// x = 3, use interpolate.
			if (mnWidth == 4) {
				nU1 = interpolate(pbYUV[1], pbYUV[5], pbYUV[5], pbYUV[5]);
				nV1 = interpolate(pbYUV[3], pbYUV[7], pbYUV[7], pbYUV[7]);
				calcYUVtoRGB(pbYUV[6], nU1, nV1, pbRGB + 9);
			}
			pbLineYUV += mnLineSize;
			pbLineRGB += nLineSizeRGB;
		}
	}
	else {
		for (int y = 0; y < mnHeight; y++) {
			const BYTE *pbYUV = pbLineYUV;
			BYTE *pbRGB = pbLineRGB;
			// x = 0
			int nU1 = interpolate(pbYUV[1], pbYUV[1], pbYUV[5], pbYUV[9]);
			int nV1 = interpolate(pbYUV[3], pbYUV[3], pbYUV[7], pbYUV[11]);
			calcYUVtoRGB(pbYUV[0], pbYUV[1], pbYUV[3], pbRGB);
			calcYUVtoRGB(pbYUV[2], nU1, nV1, pbRGB + 3);
			pbYUV += 4;
			pbRGB += 6;
			int x;
			for (x = 2; x < mnWidth - 4; x += 2) {
				nU1 = interpolate(pbYUV[-3], pbYUV[1], pbYUV[5], pbYUV[9]);
				nV1 = interpolate(pbYUV[-1], pbYUV[3], pbYUV[7], pbYUV[11]);
				calcYUVtoRGB(pbYUV[0], pbYUV[1], pbYUV[3], pbRGB);
				calcYUVtoRGB(pbYUV[2], nU1, nV1, pbRGB + 3);
				pbYUV += 4;
				pbRGB += 6;
			}
			// x = W-4
			nU1 = interpolate(pbYUV[-3], pbYUV[1], pbYUV[5], pbYUV[5]);
			nV1 = interpolate(pbYUV[-1], pbYUV[3], pbYUV[7], pbYUV[7]);
			calcYUVtoRGB(pbYUV[0], pbYUV[1], pbYUV[3], pbRGB);
			calcYUVtoRGB(pbYUV[2], nU1, nV1, pbRGB + 3);
			// x = W-2
			calcYUVtoRGB(pbYUV[4], pbYUV[5], pbYUV[7], pbRGB + 6);
			// x = W-1
			if (x + 3 < mnWidth) {
				nU1 = interpolate(pbYUV[1], pbYUV[5], pbYUV[5], pbYUV[5]);
				nV1 = interpolate(pbYUV[3], pbYUV[7], pbYUV[7], pbYUV[7]);
				calcYUVtoRGB(pbYUV[6], nU1, nV1, pbRGB + 9);
			}
			pbLineYUV += mnLineSize;
			pbLineRGB += nLineSizeRGB;
		}
	}
	return 0;
}

}	// namespace akamoz
