// $Id: vfwstream.cpp 353 2015-09-19 16:06:32Z you $
/*
 * Copyright (c) 2015 akamoz.jp
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <akamoz-windows.h>
#include <Vfw.h>
#include <map>
#include <akamoz-image.h>

namespace akamoz4 {

VfWSingleStream::VfWSingleStream() throw()
{
	akamoz::gAVIiniter.init();	// share with the first generation of akamoz
	_init();
}

template <typename TC>
int VfWSingleStream::setup(const TC *pszfn)
{
	mpavif.reset();
	if (pszfn == NULL)
		return AKAMOZ_ERR(-1);
	PAVIFILE pavif;
	if (aviFileOpen(&pavif, pszfn, OF_READ, NULL) != 0)
		return AKAMOZ_ERR(-1);
	mpavif.reset(pavif);

	AVIFILEINFO avifi;
	if (AVIFileInfo(mpavif, &avifi, sizeof(avifi)) != 0)
		return AKAMOZ_ERR(-1);
	mdwTotalFrames = avifi.dwLength;
	mnNumStreams = avifi.dwStreams;
	if (mdaStreamInfo.allocate(mnNumStreams) < 0)
		return AKAMOZ_ERR(-1);

	PAVISTREAM pavis;
	for (int i = 0; i < mnNumStreams; i++) {
		if (AVIFileGetStream(mpavif, &pavis, 0, i) != 0)
			return AKAMOZ_ERR(-1);
		AVISTREAMINFO avisi;
		if (AVIStreamInfo(pavis, &avisi, sizeof(avisi)) != 0)
			goto L1;

		AVIStreamDetail &info = mdaStreamInfo[i];
		info.fccType = avisi.fccType;
		if (info.fccType == streamtypeVIDEO)
			info.nType = aviVideo;
		else if (info.fccType == streamtypeAUDIO)
			info.nType = aviAudio;
		else if (info.fccType == streamtypeMIDI)
			info.nType = aviMidi;
		else if (info.fccType == streamtypeTEXT)
			info.nType = aviText;
		else
			info.nType = aviUnknown;
		mstrmap[info.fccType].push_back(i);
		info.dwHandler = avisi.fccHandler;
		info.dwRateNum = avisi.dwRate;
		info.dwRateDenom = avisi.dwScale;
		info.dwLength = avisi.dwLength;

		LONG fmtlen;
		if (AVIStreamFormatSize(pavis, 0, &fmtlen) != 0)
			goto L1;
		if (info.dbFormat.reuse(fmtlen) < 0)
			goto L1;
		if (AVIStreamReadFormat(pavis, 0, info.dbFormat.getmp(), &fmtlen) != 0)
			goto L1;

		AVIStreamRelease(pavis);
	}
	return 0;

L1:
	AVIStreamRelease(pavis);
	return AKAMOZ_ERR(-1);
}

template int VfWSingleStream::setup(LPCSTR pszfn);
template int VfWSingleStream::setup(LPCWSTR pszfn);


void VfWSingleStream::cleanup() throw()
{
	mdaStreamInfo.cleanup();
	mstrmap.clear();
	mpavis.reset();
	mpavif.reset();
	AVISingleStreamIF::cleanup();
	_init();
}

int VfWSingleStream::selectStream(int n, DWORD fccid)
{
	if (mpavif == NULL)
		return AKAMOZ_ERR(-1);
	mpavis.reset();
	mnStreamIndex = -1;
	PAVISTREAM pavis = NULL;
	if (AVIFileGetStream(mpavif, &pavis, fccid, n) != 0)
		return AKAMOZ_ERR(-1);

	if (fccid == 0) {
		if (n >= mnNumStreams)
			return AKAMOZ_ERR(-1);
		mnStreamIndex = n;
	}
	else {
		if (mstrmap.count(fccid) < 1)
			return AKAMOZ_ERR(-1);
		if ((UINT)n >= mstrmap[fccid].size())
			return AKAMOZ_ERR(-1);
		mnStreamIndex = mstrmap[fccid][n];
	}

	mpavis.reset(pavis);
	mnFrameCnt = 0;
	return 0;
}

int VfWSingleStream::getNextFrame(DataBuffer *pdb)
{
	if (mpavis == NULL)
		return AKAMOZ_ERR(-1);
	if (mnFrameCnt >= AVIStreamLength(mpavis))
		return 0;
	LONG sz, len;
	if (AVIStreamSampleSize(mpavis, mnFrameCnt, &sz) != 0)
		return AKAMOZ_ERR(-1);
	if (pdb->reuse(sz) < 0)
		return AKAMOZ_ERR(-1);
	LPVOID pv = pdb->getmp();
	if (pv == NULL)
		return AKAMOZ_ERR(-1);
	if (AVIStreamRead(mpavis, mnFrameCnt, 1, pv, sz, NULL, &len) != 0)
		return AKAMOZ_ERR(-1);
	if (len != 1)
		return AKAMOZ_ERR(-1);
	mnFrameCnt++;
	return 1;
}

} // namespace akamoz4
