// $Id: akamoz-image.h 353 2015-09-19 16:06:32Z you $
/*
 * Copyright (c) 2007-2015 akamoz.jp
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __AKAMOZ_IMAGE_INCLUDED
#define __AKAMOZ_IMAGE_INCLUDED

#include <akamoz-data.h>
#include <stdio.h>	// for FILE *
#include <vector>
#include <map>

#ifdef AKAMOZ_WIN32
#include <windows.h>
#include <vfw.h>
#endif

#include <akamoz-fourcc.h>

enum AVIStreamType {
	aviUnknown, aviAudio, aviMidi, aviText, aviVideo
};

struct AVIStreamSummary {
	int nType;
	DWORD dwHandler;
	DWORD dwRateNum;	// t[[gqB
	DWORD dwRateDenom;	// t[[gB
	DWORD dwLength;	// t[邢̓TvB
	DWORD fccType;	// Xg[^Cv
};

namespace akamoz {

class ImageBuffer {
  public:
	ImageBuffer();
	virtual ~ImageBuffer();
	int setProperty(int nWidth, int nHeight, int nGuard = -1, int nAlign = -1);
	inline int setProperty(const ImageBuffer *pimg, int nGuard = -1, int nAlign = -1) {
		return setProperty(pimg->getWidth(), pimg->getHeight(), nGuard, nAlign);
	}
	inline int setProperty(const ImageBuffer &rimg, int nGuard = -1, int nAlign = -1) {
		return setProperty(&rimg, nGuard, nAlign);
	}
	BYTE *getPixelBuffer() { return mpbBuf; }
	const BYTE *getPixelBuffer() const { return mpbBuf; }
	int getImageSize() const { return mnImageSize; }
	int getLineSize() const { return mnLineSize; }
	int getWidth() const { return mnWidth; }
	int getHeight() const { return mnHeight; }
	DWORD getFormat() const { return fccFormat; }
	void setAlign(int nAlign) { mnAlign = nAlign > 0 ? nAlign : 4; }
	void setGuardBytes(int nGuard) { mnGuardBytes = nGuard > 0 ? nGuard : 0; }
	int cleanup();
	inline bool isReady() const {
		return (mnWidth > 0) && (mnHeight > 0) && (mnLineSize > 0) &&
			(mnImageSize > 0) && (mdabBuf != NULL);
	}
	inline bool isClean() const {
		return (mnWidth < 0) && (mnHeight < 0) && (mnLineSize < 0) &&
			(mnImageSize < 0) && (mdabBuf == NULL);
	}
  protected:
	BYTE *mpbBuf;
	int mnWidth;
	int mnHeight;
	int mnLineSize;
	int mnImageSize;
	int mnAlign;
	int mnGuardBytes;
	DWORD fccFormat;
	virtual int _calcLineSize(int nWidth);
  private:
	darray<BYTE> mdabBuf;
	void _init() {
		mnWidth = -1;
		mnHeight = -1;
		mnLineSize = -1;
		mnImageSize = -1;
		fccFormat = 0;	// BI_RGB
		mpbBuf = 0;
		mnGuardBytes = 0;
		mnAlign = 4;
	}
};

template <class TIMG, typename TPIXPTR>
class TImageItorBase {
  public:
	typedef TIMG TImage;
	typedef TPIXPTR TPixPtr;
	typedef TImageItorBase<TIMG, TPIXPTR> TThis;

	inline TImageItorBase() { _init(); }
	inline TImageItorBase(TIMG &rimg, int nOffset = 0) { _init(); setup(rimg, nOffset); }
	inline TImageItorBase(TIMG *pimg, int nOffset = 0) { _init(); setup(pimg, nOffset); }
	inline int setup(TIMG &rimg, int nOffset = 0) { return setup(&rimg, nOffset); }
	int setup(TIMG *pimg, int nOffset = 0);
	inline TThis &operator ++ () { _inc(); return *this; }
	inline TThis operator ++ (int) { TThis ret = *this; _inc(); return ret; }
	inline TThis &operator += (int n) { mpbPix += n * mnLineSize; return *this; }
	inline TThis operator + (int n) const { TThis ret = *this; return ret += n; }
	inline TThis &operator -- () { _dec(); return *this; }
	inline TThis operator -- (int) { TThis ret = *this; _dec(); return ret; }
	inline TThis operator -= (int n) { mpbPix -= n * mnLineSize; return *this; }
	inline TThis operator - (int n) const { TThis ret = *this; return ret -= n; }
	inline TPIXPTR operator *() const { return mpbPix; }
	inline int getlnsz() const { return mnLineSize; }
  protected:
	TPIXPTR mpbPix;
	int mnLineSize;
  private:
	void _init() {
		mpbPix = NULL;
		mnLineSize = 0;
	}
	inline void _inc() { mpbPix += mnLineSize; }
	inline void _dec() { mpbPix -= mnLineSize; }
};

typedef TImageItorBase<ImageBuffer, BYTE *> ImageItor;

class ImageItorConst : public TImageItorBase<const ImageBuffer, const BYTE *> {
  public:
	typedef TImageItorBase<const ImageBuffer, const BYTE *> TBase;

	inline ImageItorConst() { }
	inline ImageItorConst(const ImageItor &imgitr) { *this = imgitr; }
	inline ImageItorConst(TImage &rimg, int nOffset = 0) : TBase(rimg, nOffset) {	}
	inline ImageItorConst(TImage *pimg, int nOffset = 0) : TBase(pimg, nOffset) { }
	ImageItorConst &operator = (const ImageItor &itor) {
		mpbPix = *itor;
		mnLineSize = itor.getlnsz();
		return *this;
	}
	ImageItorConst(const TBase &itor) {
		mpbPix = *itor;
		mnLineSize = itor.getlnsz();
	}
};

template <class TIMGITR, typename TDEREF, int BYTES>
class TPixelItorBase {
  public:
	typedef TIMGITR TImageItor;
	typedef TPixelItorBase<TIMGITR, TDEREF, BYTES> TThis;
	typedef typename TIMGITR::TPixPtr TPixPtr;

	inline TPixelItorBase() { _init(); }
	inline TPixelItorBase(TIMGITR &rimgitor, int nOffset = 0) { _init(); setup(rimgitor, nOffset); }
	inline TPixelItorBase(TIMGITR *pimgitor, int nOffset = 0) { _init(); setup(pimgitor, nOffset); }
	inline int setup(TIMGITR &rimgitor, int nOffset = 0) { return setup(&rimgitor, nOffset); }
	int setup(TIMGITR *pimgitor, int nOffset = 0) {
		if (pimgitor == NULL)
			return AKAMOZ_ERR(-1);
		mpbPix = **pimgitor;
		if (mpbPix == NULL)
			return AKAMOZ_ERR(-1);
		mpbPix += nOffset * BYTES;
		return 0;
	}
	inline TThis &operator ++ () { _inc(); return *this; }
	inline TThis operator ++ (int) { TThis ret = *this; _inc(); return ret; }
	inline TThis &operator += (int n) { mpbPix += n * BYTES; return *this; }
	inline TThis operator + (int n) const { TThis ret = *this; return ret += n; }
	inline TThis &operator -- () { _dec(); return *this; }
	inline TThis operator --(int) { TThis ret = *this; _dec(); return ret; }
	inline TThis &operator -= (int n) { mpbPix -= n * BYTES; return *this; }
	inline TThis operator - (int n) const { TThis ret = *this; return ret -= n; }
	inline TThis operator () (int n) const { return *this + n; }
	inline TDEREF operator[] (int n) const { return this->mpbPix[n]; }
	inline TPixPtr getptr() const { return mpbPix; }
  protected:
	TPixPtr mpbPix;
  private:
	void _init() { mpbPix = NULL; }
	inline void _inc() { mpbPix += BYTES; }
	inline void _dec() { mpbPix -= BYTES; }
};

template <int NBYTES>
class PixelItor : public TPixelItorBase<ImageItor, BYTE &, NBYTES> {
  public:
	typedef TPixelItorBase<ImageItor, BYTE &, NBYTES> TBase;
	typedef typename TBase::TImageItor TImageItor;
	inline PixelItor() { }
	inline PixelItor(TImageItor &rimgitr, int nOffset = 0) : TBase(rimgitr, nOffset) { }
	inline PixelItor(TImageItor *pimgitr, int nOffset = 0) : TBase(pimgitr, nOffset) { }
};

template <int NBYTES>
class PixelItorConst : public TPixelItorBase<ImageItorConst, BYTE, NBYTES> {
  public:
	typedef TPixelItorBase<ImageItorConst, BYTE, NBYTES> TBase;
	typedef typename TBase::TImageItor TImageItor;
	typedef PixelItorConst<NBYTES> TThis;
	inline PixelItorConst(const PixelItor<NBYTES> &pxitr) { *this = pxitr; }
	inline PixelItorConst() { }
	inline PixelItorConst(TImageItor &rimgitr, int nOffset = 0) : TBase(rimgitr, nOffset) { }
	inline PixelItorConst(TImageItor *pimgitr, int nOffset = 0) : TBase(pimgitr, nOffset) { }
	inline TThis & operator = (const PixelItor<NBYTES> &pxitr) { this->mpbPix = pxitr.getptr(); return *this; }
};

class ImageBufferYUY2 : public ImageBuffer {
  public:
	ImageBufferYUY2();
	int setimg(const ImageBuffer *pimg);
	int setimg(const ImageBuffer &rimg) {
		return setimg(&rimg);
	}
	static inline int calcY(BYTE bRed, BYTE bGreen, BYTE bBlue) {
		// 210300 = 11*500 + 16*12800
		return (11*299*bRed + 11*587*bGreen + 11*114*bBlue + 210300) / 12800;
	}
	static inline int calcU(BYTE bRed, BYTE bGreen, BYTE bBlue) {
		// 58291712 = 128.5 * 256 * 1772
		return (-225*299*bRed -225*587*bGreen +225*886*bBlue + 58291712)/(256*1772);
	}
	static inline int calcV(BYTE bRed, BYTE bGreen, BYTE bBlue) {
		// 46120192 = 128.5 * 256 * 1402
		return (225*701*bRed -225*587*bGreen -225*114*bBlue + 46120192)/(256*1402);
	}
	static void calcYUVtoRGB(int nY, int nU, int nV, BYTE *pRGB);
	int convertToRGB(ImageBuffer *pimgRGB) const;
	inline int convertToRGB(ImageBuffer &imgRGB) const {
		return convertToRGB(&imgRGB);
	}
  protected:
	virtual int _calcLineSize(int nWidth);
	static inline int interpolate(int A, int B, int C, int D) {
		return ((9 * (B + C)) - A - D + 8) >> 4;
	}
	static inline int clip256(int n) {
		if (n < 0)
			return 0;
		if (n > 255)
			return 255;
		return n;
	}
  private:
	void _init() {
		fccFormat = AKAMOZ_FOURSTR("YUY2");
	}
};

int readBitmapFile(ImageBuffer *pimg, const char *fn);
int writeBitmapFile(const ImageBuffer *pimg, FILE *fp);
int writeBitmapFile(const ImageBuffer *pimg, const char *fn);

#ifdef AKAMOZ_HAVE_WIN32VFW

inline HRESULT aviStreamOpenFromFile(
   PAVISTREAM *ppavi, LPCSTR szFile, DWORD fccType, LONG lParam,
   UINT mode, CLSID *pHdlr
) {
	return ::AVIStreamOpenFromFileA(
		ppavi, szFile, fccType, lParam, mode, pHdlr
	);
}

inline HRESULT aviStreamOpenFromFile(
   PAVISTREAM *ppavi, LPCWSTR szFile, DWORD fccType, LONG lParam,
   UINT mode, CLSID *pHdlr
) {
	return ::AVIStreamOpenFromFileW(
		ppavi, szFile, fccType, lParam, mode, pHdlr
	);
}

class AVIStreamInitializer {
  public:
	AVIStreamInitializer();
	~AVIStreamInitializer();
	void init();
};

extern AVIStreamInitializer gAVIiniter;

class AVIStream {
  public:
	AVIStream();
	template <typename TC>
	AVIStream(const TC *pszfn);
	~AVIStream();
	template <typename TC>
	int setup(const TC *pszfn);
	int cleanup();
	int getInfo(AVISTREAMINFO *pavisi);
	int getFormat(BITMAPINFOHEADER *pbmih, long lPos = 0);
	int getFormat(DataBuffer *pdb, long lPos = 0);
	int getFrame(ImageBuffer *pimg, long lPos);
	int getFrame(DataBuffer *pdb, long lPos);
	int getFrameSize(long lPos);
	int isReady() const { return mbOpened; }
	int isClean() const { return !mbOpened; }
  private:
	UINT mbOpened:1;
	PAVISTREAM mpavistr;
	void _init();
	void _cleanup();
};

#endif // #ifdef AKAMOZ_HAVE_WIN32VFW

} // namespace akamoz

namespace akamoz2 {

typedef ::AVIStreamSummary AVIStreamSummary;

//--------------------------------------------------------------------
// I M A G E   B U F F E R S
//--------------------------------------------------------------------
class ImageBuffer {	// akamoz2
  public:
	ImageBuffer() throw() { _init(); }
	virtual ~ImageBuffer();
	int setup(int x, int y, int bpp, BYTE *pbBuf = AKAMOZ_NULL) throw();
	int setup(int x, int y, int bpp, const BYTE *pbBuf) throw();
	BYTE *getmp() throw() { return mdb.getmp(); }
	const BYTE *getcp() const throw() { return mdb.getcp(); }
	void getptr(BYTE *&rpb) { rpb = getmp(); }
	void getptr(const BYTE *&rpb) const { rpb = getcp(); }
	int getWidth() const throw() { return mnWidth; }
	int getHeight() const throw() { return mnHeight; }
	int getBPP() const throw() { return msbBPP; }
	int getPackSize() const throw() { return msbPackSize; }
	int getPixPacked() const throw() { return msbPixPacked; }
	int getLineSize() const throw() { return mnLineSize; }
	int getLineSpan() const throw() { return mbBottomUp ? -mnLineSize : mnLineSize; }
	int getImageSize() const throw() { return mnImageSize; }
	void setBottomUp(int b = 1) throw() { mbBottomUp = b; }
	void setTopDown(int b = 1) throw() { mbBottomUp = !b; }
	int isBottomUp() const throw() { return mbBottomUp; }
	int isTopDown() const throw() { return !mbBottomUp; }
  protected:
	virtual int _setup();
	virtual int _setPacking();
	virtual int _calcLineSize();
	virtual int _calcImageSize();
	virtual int _setOptions();
	virtual int _postsetup();
	DataBuffer mdb;
	int mnWidth;
	int mnHeight;
	int mnLineSize;
	int mnImageSize;
	SBYTE msbBPP;
	SBYTE msbPackSize;
	SBYTE msbPixPacked;
	UINT mbBottomUp:1;
  private:
	void _init() throw() {
		mnWidth = mnHeight = mnLineSize = mnImageSize = -1;
		msbBPP = msbPackSize = msbPixPacked = -1;
		mbBottomUp = 0;
	}
	int setup_base(int x, int y, int bpp) throw();
};

class ImageBufferYUV : public ImageBuffer {
  protected:
	class Plane : public ImageBuffer {
	  public:
		int setPixSpan(int nPixSpan);
	};
	Plane mimgY;
	Plane mimgU;
	Plane mimgV;
  public:
	ImageBuffer &Y() throw() { return mimgY; }
	ImageBuffer &U() throw() { return mimgU; }
	ImageBuffer &V() throw() { return mimgV; }
	const ImageBuffer &Y() const throw() { return mimgY; }
	const ImageBuffer &U() const throw() { return mimgU; }
	const ImageBuffer &V() const throw() { return mimgV; }
};

// TIMP must have ImageBufferYUV as the base class, 
// must have _setupPlanes(const BYTE *),
// and can optionally have _setupPlanes(BYTE *)
template <class TIMP>
class ImageBufferYUVImp : public TIMP {
  protected:
	virtual int _postsetup() {
		BYTE *pbMutable = this->getmp();
		if (pbMutable != NULL)
			return this->_setupPlanes(pbMutable);
		const BYTE *pbConst = this->getcp();
		if (pbConst == NULL)
			return AKAMOZ_ERR(-1);
		return this->_setupPlanes(pbConst);
	}
};

class ImageBufferYUY2Imp : public ImageBufferYUV {
  protected:
	template <typename T> int _setupPlanes(T *p);
};

typedef class ImageBufferYUVImp<ImageBufferYUY2Imp> ImageBufferYUY2;

class ImageBufferYV12Imp : public ImageBufferYUV {
  public:
	int setup(const ImageBufferYUY2 *pimg, int bInterlaced = 0);
  protected:
	template <typename T> int _setupPlanes(T *p);
  private:
	void convLuma(const ImageBuffer *psrc, ImageBuffer *pdst);
	void convChroma2P(const ImageBuffer *psrc, ImageBuffer *pdst);
	void convChroma2I(const ImageBuffer *psrc, ImageBuffer *pdst);
};

typedef class ImageBufferYUVImp<ImageBufferYV12Imp> ImageBufferYV12;

//--------------------------------------------------------------------
// I M A G E   I T E R A T O R S
//--------------------------------------------------------------------
template <class TIMG, typename TPPIX>
class ImageItorBase {
  public:
	typedef ImageItorBase<TIMG, TPPIX> TThis;
	int setup(TIMG *pimg, int y = 0, int x = 0, int offset = 0) throw() {
		AKAMOZ_ASSUME(pimg != NULL);
		AKAMOZ_ASSUME(pimg->getHeight() > 0);
		AKAMOZ_ASSUME(pimg->getPackSize() > 0);
		AKAMOZ_ASSUME((x == 0) || (pimg->getPixPacked() == 1));
		pimg->getptr(ppix);
		AKAMOZ_ASSUME(ppix != NULL);
		mnLineSpan = pimg->getLineSpan();
		if (mnLineSpan < 0)
			ppix -= mnLineSpan * (pimg->getHeight() - 1);
		ppix += mnLineSpan * y;
		ppix += pimg->getPackSize() * x;
		ppix += offset;
		return 0;
	}
	int setup(TIMG &pimg, int y = 0, int x = 0, int offset = 0) throw() {
		return setup(&pimg, y, x, offset);
	}
	TPPIX get() const throw() { return ppix; }
	TPPIX operator * () const throw() { return ppix; }
	TThis &operator ++ () { _inc(); return *this; }
	TThis operator ++ (int) { TThis ret = *this; _inc(); return ret; }
	TThis &operator += (int n) { ppix += n * mnLineSpan; return *this; }
	TThis operator + (int n) { TThis ret = *this; ret += n; return ret; }
  protected:
	TPPIX ppix;
	int mnLineSpan;
	void _inc() { ppix += mnLineSpan; }
	void _dec() { ppix -= mnLineSpan; }
};

class ImageItor
	: public ImageItorBase<ImageBuffer, BYTE *>
{
  public:
	ImageItor(ImageBuffer *pimg, int y = 0, int x = 0, int offset = 0) {
		setup(pimg, y, x, offset);
	}
};

class ImageItorConst
	: public ImageItorBase<const ImageBuffer, const BYTE *>
{
  public:
	ImageItorConst() { }
	ImageItorConst(const ImageBuffer *pimg, int y = 0, int x = 0, int offset = 0) {
		setup(pimg, y, x, offset);
	}
};

class VideoStream {
  public:
	virtual ~VideoStream();
	virtual int read(ImageBuffer *pimg) = 0;
};

class VideoStreamAVI : public VideoStream {
  public:
	VideoStreamAVI() throw() { _init(); }
//	int setup(AVIStream &pavi, int nStream) throw();
  private:
	void _init() throw() {
	}
};

class AVIStream {	// akamoz2
  public:
	AVIStream() throw();
	~AVIStream() throw();
	template <typename TC>
	int setup(const TC *pszfn) throw();
	int getNumStreams() { return mnNumStreams; }
	int getStreamSummary(int nStr, AVIStreamSummary &info);
	const DataBuffer *getStreamDetail(int nStr);
	int getVideoStream(int nStr = 0);
	int getNextStrnum();
	int nextFrame();
	int getFrame(DataBuffer *pdb);
  private:
	struct AVIStreamDetail : public AVIStreamSummary {
		DataBuffer dbFormat;
	};
	ReadLargeFile mrf;
	ReadLargeIFF miffRIFF;
	ReadLargeIFF miffAVI;
	ReadLargeIFF miffMovi;
	darray<AVIStreamDetail> mdaStreamInfo;
	int mnNumStreams;
	int mnLoopStat;
	void _init() {
		mnNumStreams = 0;
		mnLoopStat = 0;
	}
	int readAVIMainHeader(ReadLargeFileGeneric &rf);
	int readAVIStreamHeader(ReadLargeFileGeneric &rf, AVIStreamDetail &info);
};

#ifdef AKAMOZ_HAVE_WIN32VFW
class vcmDecode {
  public:
	vcmDecode() { _init(); }
	virtual ~vcmDecode();
	int setup(const AVIStreamSummary &aviinfo, const BITMAPINFOHEADER *pbmih, 
		DWORD fccTarget = 0xffffffff, int bpp = 24);
	int getFormat(BITMAPINFOHEADER &rbmih) { rbmih = mbmihDst; return 0; }
	int decode(DataBuffer *pdbSrc, ImageBuffer *pimgDst);
	int cleanup();
	int getInfo(ICINFO &info);
  private:
	HIC mhic;
	darray<BYTE> mdabmihSrc;
	BITMAPINFOHEADER mbmihDst;
	UINT mbBegun:1;
	void _init() {
		mhic = NULL;
		mbBegun = 0;
	}
};
#endif // #ifdef AKAMOZ_HAVE_WIN32VFW

} // namespace akamoz2

#include <map>

namespace akamoz4 {

class AVISingleStreamIF {	// akamoz4
  public:
	AVISingleStreamIF() { _init(); }

	// virtual functions
	virtual ~AVISingleStreamIF();
	virtual void cleanup() throw();
	virtual int selectStream(int n, DWORD fccid = 0) = 0;
	virtual int getNextFrame(DataBuffer *pdb) = 0;

	// inline utility
	int getNumStreams() const { return mnNumStreams; }
	int getStreamIndex() const { return mnStreamIndex; }
	DWORD getOverallNumFrames() { return mdwTotalFrames; }

	// non-inline utility
	int getNumStreams(DWORD fccid) const;
	int selectVideoStream(int n);
	int selectAudioStream(int n);
	int selectMidiStream(int n);
	int selectTextStream(int n);
	int getNumVideoStreams() const;
	int getNumAudioStreams() const;
	int getNumMidiStreams() const;
	int getNumTextStreams() const;
	int getStreamSummary( AVIStreamSummary &info, int nStr = -1) const;
	const DataBuffer *getStreamDetail(int nStr = -1) const;
	const DataBuffer *getStreamSupplement(int nStr = -1) const;
  protected:
	struct AVIStreamDetail : public AVIStreamSummary {
		DataBuffer dbFormat;
	};
	akamoz2::darray<AVIStreamDetail> mdaStreamInfo;
	std::map<DWORD, std::vector<int> > mstrmap;
	int mnNumStreams;
	int mnStreamIndex;
	DWORD mdwTotalFrames;	// from AVIMainHeader, or OpenDML Header
	void _init() {
		mnNumStreams = 0;
		mdwTotalFrames = 0;
		mnStreamIndex = -1;
	}
};

template <typename TRF, typename TEND = LittleEndian>
class AVISingleStreamBase : public AVISingleStreamIF {	// akamoz4
  public:
	typedef TRF TReadFile;
	typedef TEND TEndian;
	typedef ChunkReader<TReadFile, TEndian> TChunkReader;
	AVISingleStreamBase() throw();
	template <typename TC>
	int setup(const TC *pszfn);
	virtual void cleanup() throw();
	virtual int selectStream(int n, DWORD fccid = 0);
	virtual int getNextFrame(DataBuffer *pdb);
#if 0
	void dumpckcnt();
#endif
  private:
	TReadFile mrf;
	TChunkReader mckRIFF;
	TChunkReader mckMovi;
	TChunkReader mckRec;
	BYTE mbFccPrefix[2];
	UINT mbRecChunk:1;
#if 0
	std::map<DWORD, int> mckcnt;
#endif
	void _init() {
		mbRecChunk = 0;
	}
	int setupAVIFile(TReadFile *prf) throw();
	int readAVIMainHeader(TChunkReader &rf);
	int readAVIStreamHeader(TChunkReader &rf, AVIStreamDetail &info);
	int checkStreamPrefix(const TChunkReader &ck);
#if 0
	int riffloop();
	int moviloop();
	int recloop();
	int streamloop();
#endif
	int readdata(DataBuffer *pdb, const TChunkReader &ck);
};

#ifdef AKAMOZ_HAVE_READLARGEFILE
#define AKAMOZ_HAVE_AVISINGLESTREAM
typedef AVISingleStreamBase<akamoz2::ReadLargeFile> AVISingleStream;
#endif

#ifdef AKAMOZ_HAVE_WIN32VFW

inline HRESULT aviFileOpen(
	PAVIFILE *ppfile, LPCSTR szFile, UINT mode, LPCLSID pclsidHandler
) {
	return ::AVIFileOpenA(ppfile, szFile, mode, pclsidHandler);
}

inline HRESULT aviFileOpen(
	PAVIFILE *ppfile, LPCWSTR szFile, UINT mode, LPCLSID pclsidHandler
) {
	return ::AVIFileOpenW(ppfile, szFile, mode, pclsidHandler);
}

class AVIFileDeleter {
  public:
	void operator () (PAVIFILE p) { AVIFileRelease(p); }
};

class AVIStreamDeleter {
  public:
	void operator () (PAVISTREAM p) { AVIStreamRelease(p); }
};

class VfWSingleStream : public AVISingleStreamIF {	// akamoz4
  public:
	VfWSingleStream() throw();
	template <typename TC>
	int setup(const TC *pszfn);
	virtual void cleanup() throw();
	virtual int selectStream(int n, DWORD fccid = 0);
	virtual int getNextFrame(DataBuffer *pdb);
  private:
	tiny_ptr<rmptr<PAVIFILE>::type, AVIFileDeleter> mpavif;
	tiny_ptr<rmptr<PAVISTREAM>::type, AVIStreamDeleter> mpavis;
	LONG mnFrameCnt;
	void _init() {
		mnFrameCnt = 0;
	}
};
#endif // #ifdef AKAMOZ_HAVE_WIN32VFW

class AVISingleStreamProxy : public tiny_ptr<AVISingleStreamIF> {
  public:
	template <typename TC>
	int setup(const TC *fname);
};

} // namespace akamoz4

#endif
