// $Id: dump-yuv4mpeg2.cpp 353 2015-09-19 16:06:32Z you $
/*
 * Copyright (c) 2007-2015 akamoz.jp
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <io.h>
#include <fcntl.h>
#include <windows.h>
#include <iostream>
#include <iomanip>
#include <akamoz-data.h>
#include <akamoz-image.h>
#include <akamoz-util.h>
#include <akamoz-iomanip.h>

#ifndef DUMP_YUV4MPEG2_PROGNAME
#define DUMP_YUV4MPEG2_PROGNAME	"dump-yuv4mpeg2"
#endif
#define DUMP_YUV4MPEG2_VERSION	"0.1"
#define MARK	(std::cerr << __LINE__ << " " << std::flush)

void dumpchar(int c)
{
	if (c < 0x20) c = '.';
	if (c >= 0x7f) c = '.';
	std::cerr << (char)c;
}

void dumpFourCC(akamoz2::DWORD dw)
{
	dumpchar(dw & 0xff);
	dumpchar((dw >> 8) & 0xff);
	dumpchar((dw >> 16) & 0xff);
	dumpchar((dw >> 24) & 0xff);
}

class ParseOption : public akamoz2::ParseOption {
  public:
	enum EOptid {
		optInterlaced, optDenominator, optPixelAspect, optUsage, optVersion
	};
	akamoz::UINT mnDenom;
	akamoz::UINT mnParNumr;
	akamoz::UINT mnParDenom;
	akamoz::UINT mbInterlaced:1;
	akamoz::UINT mbVersion:1;
	ParseOption() {
		mbInterlaced = 0;
		mnDenom = 0;
		mnParNumr = mnParDenom = 0;
	}
	int parse(int argc, akamoz2::CCHAR * const argv[]) {
		return akamoz2::ParseOption::parse(argc, argv, moptinfo);
	}
  protected:
	static akamoz2::ParseOptionInfo moptinfo[];
	virtual int proc(const akamoz2::ParseOptionInfo *pinfo, akamoz2::CCHAR *parg);
};

struct akamoz2::ParseOptionInfo ParseOption::moptinfo[] = {
	{ "i", 0, optInterlaced, "interlaced", 0 },
	{ "par", 1, optPixelAspect, "pixel aspect ratio (w:h)", 0 },
	{ "denom", 1, optDenominator, "specify denominator for framerate", 0 },
	{ "version", 0, optVersion, "print version and exit", 0 },
	{ "h", 0, optUsage, "print this message", 0 },
	{ "?", 0, optUsage, "same as -h", 0 },
	{ 0, 0, 0, 0, 0 }
};

int ParseOption::proc(const akamoz2::ParseOptionInfo *pinfo, akamoz2::CCHAR *s)
{
	switch (pinfo->nID) {
	  case optInterlaced:
		mbInterlaced = 1;
		break;
	  case optDenominator:
		s = akamoz2::getuint(s, mnDenom);
		if ((s == NULL) || (*s != '\0'))
			return AKAMOZ_ERR(-1);
		break;
	  case optPixelAspect:
		s = akamoz2::getuint(s, mnParNumr);
		s = akamoz2::checkdelim(s, ':');
		s = akamoz2::getuint(s, mnParDenom);
		if ((s == NULL) || (*s != '\0'))
			return AKAMOZ_ERR(-1);
		break;
	  case optVersion:
		mbVersion = 1;
		return 1;
	  case optUsage:
		usage(std::cerr, moptinfo);
		return AKAMOZ_ERR(-1);
	  default:
		return AKAMOZ_ERR(-1);
	}
	return 0;
}

void dumpYUV4MPEG2Header(
	const BITMAPINFOHEADER *pbmih,
	const akamoz2::AVIStreamSummary &info,
	const ParseOption &opt
) {
	UINT numr, denom;
	if (opt.mnDenom == 0) {
		numr = (UINT)info.dwRateNum;
		denom = (UINT)info.dwRateDenom;
	}
	else {
		float fps = (float)info.dwRateNum / info.dwRateDenom;
		numr = (int)(fps * opt.mnDenom + 0.5);
		denom = opt.mnDenom;
	}
	printf("YUV4MPEG2 W%u H%u C420mpeg2 I%c F%u:%u",
		(UINT)pbmih->biWidth, (UINT)pbmih->biHeight,
		opt.mbInterlaced ? 't' : 'p', numr, denom);
	if (opt.mnParDenom != 0)
		printf(" A%u:%u", (UINT)opt.mnParNumr, (UINT)opt.mnParDenom);
	printf("\n");
}

void dumpYUV4MPEG2Frame(const akamoz2::ImageBufferYV12 &img)
{
	printf("FRAME\n");
	fwrite(img.Y().getcp(), 1, img.Y().getImageSize(), stdout);
	fwrite(img.U().getcp(), 1, img.U().getImageSize(), stdout);
	fwrite(img.V().getcp(), 1, img.V().getImageSize(), stdout);
}

int main(int argc, const char *const argv[])
{
	ParseOption opt;
	if (opt.parse(argc - 1, argv + 1) < 0)
		return AKAMOZ_ERR(1);
	if (opt.mbVersion) {
		std::cerr << DUMP_YUV4MPEG2_PROGNAME << ": version " <<
			DUMP_YUV4MPEG2_VERSION << "\n";
		return 0;
	}
	akamoz2::CCHAR *pszInfile = opt.getarg();
	if (pszInfile == NULL)
		return AKAMOZ_ERR(2);

	akamoz4::AVISingleStreamProxy pavi;
	if (pavi.setup(pszInfile) < 0)
		return AKAMOZ_ERR(10);
	std::cerr << "Number of frames from overall header: " <<
		pavi->getOverallNumFrames() << "\n";

	if (pavi->selectVideoStream(0) < 0)
		return AKAMOZ_ERR(40);
	std::cerr << "The first video stream is: " << pavi->getStreamIndex() << "\n";

	akamoz2::AVIStreamSummary info;
	if (pavi->getStreamSummary(info) < 0)
		return AKAMOZ_ERR(20);
	std::cerr << "Length=" << info.dwLength << ", rate=" <<
		info.dwRateNum / (double)info.dwRateDenom <<
		", handler: ";
	dumpFourCC(info.dwHandler);
	std::cerr << "\n";

	const akamoz2::DataBuffer *pdbFormat = pavi->getStreamDetail();
	if (pdbFormat == NULL)
		return AKAMOZ_ERR(21);
	if (pdbFormat->size() < (int)sizeof(BITMAPINFOHEADER))
		return AKAMOZ_ERR(22);
	std::cerr << "Size of stream detail: " << pdbFormat->size() << "\n";
	const BITMAPINFOHEADER *pbmih = (const BITMAPINFOHEADER *)pdbFormat->getcp();
	if (pbmih == NULL)
		return AKAMOZ_ERR(22);
	std::cerr << "Width: " << pbmih->biWidth << ", Height: " <<
		pbmih->biHeight << ", " << pbmih->biBitCount << "bpp, compression: ";
	dumpFourCC(pbmih->biCompression);
	std::cerr << ", biSize: " << pbmih->biSize;
	std::cerr << "\n";

	akamoz2::vcmDecode dec;
	if (dec.setup(info, pbmih, akamoz2::convfccdw("YUY2"), 16) < 0)
		return AKAMOZ_ERR(30);
	BITMAPINFOHEADER bmihDst;
	if (dec.getFormat(bmihDst) < 0)
		return AKAMOZ_ERR(32);
	std::cerr << "Output Format: ";
	dumpFourCC(bmihDst.biCompression);
	ICINFO icinfo;
	if (dec.getInfo(icinfo) < 0)
		return AKAMOZ_ERR(41);
	std::cerr << ", Handler: ";
	dumpFourCC(icinfo.fccHandler);
	std::cerr << "\n";

	akamoz2::DataBuffer dbimg;
	akamoz2::ImageBufferYUY2 imgDst;
#ifdef AKAMOZ_WIN32
	setmode(fileno(stdout), O_BINARY);
#endif
	dumpYUV4MPEG2Header(pbmih, info, opt);
	akamoz2::ImageBufferYV12 imgYV12;
	int nFrm = 0;
	for (;;) {
		int r;
		if ((r = pavi->getNextFrame(&dbimg)) < 0)
			return AKAMOZ_ERR(31);
		if (r == 0)
			break;
		std::cerr << nFrm++ << "\r";
		if (dec.decode(&dbimg, &imgDst) < 0)
				return AKAMOZ_ERR(33);
		if (imgYV12.setup(&imgDst, opt.mbInterlaced) < 0)
			return AKAMOZ_ERR(34);
		dumpYUV4MPEG2Frame(imgYV12);
	}
	std::cerr << nFrm << "\n";

	return 0;
}
